"""
Module for Terraform Cloud API Endpoint: Variable Sets.
"""

from .endpoint import TFCEndpoint


class TFCVarSets(TFCEndpoint):
    """
    `Variable Sets API Docs \
        <https://www.terraform.io/docs/cloud/api/variable-sets.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._org_varsets_base_url = f"{self._api_v2_base_url}/organizations/{self._org_name}/varsets"
        self._endpoint_base_url = f"{self._api_v2_base_url}/varsets"
        self._ws_api_v2_base_url = f"{self._api_v2_base_url}/workspaces"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, payload):
        """
        ``POST organizations/:organization_name/varsets``

        `Variable Sets Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#create-a-variable-set>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#sample-payload>`_
        """
        return self._create(self._org_varsets_base_url, payload)

    def show(self, varset_id):
        """
        ``GET varsets/:varset_id``

        `Variable Sets Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#show-variable-set>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}"
        return self._show(url)

    def list_for_org(self):
        """
        ``GET organizations/:organization_name/varsets``

        `Variable Sets List For Organization API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#list-variable-set>`_
        """
        return self._list(self._org_varsets_base_url)

    def list_all_for_org(self):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every workspace in an organization.

        Returns an object with two arrays of objects.
        """
        return self._list_all(self._org_varsets_base_url)

    def list_for_workspace(self, workspace_id):
        """
        ``GET workspaces/:workspace_id/varsets``

        `Variable Sets List For Workspace API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#list-variable-set>`_
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/varsets"
        return self._list(url)

    def list_all_for_workspace(self, workspace_id):
        """
        ``GET workspaces/:workspace_id/varsets``

        `Variable Sets List For Workspace API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#list-variable-set>`_
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/varsets"
        return self._list_all(url)

    def update(self, varset_id, payload):
        """
        ``PUT/PATCH varsets/:varset_id``

        `Variable Sets Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#update-a-variable-set>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#sample-payload-1>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}"
        return self._update(url, payload)

    def destroy(self, varset_id):
        """
        ``DELETE varsets/:varset_id``

        `Variable Sets Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#delete-a-variable-set>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}"
        return self._destroy(url)

    # VARIABLE RELATIONSHIPS
    def add_var_to_varset(self, varset_id, payload):
        """
        ``POST varsets/:varset_external_id/relationships/vars``

        `Variable Sets Add Variable API Doc Reference \
            <POST varsets/:varset_external_id/relationships/vars>`_

        `Add Variable To Variable Set Sample Payload \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#sample-payload-2>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}/relationships/vars"
        return self._post(url, data=payload)

    def update_var_in_varset(self, varset_id, var_id, payload):
        """
        ``PATCH varsets/:varset_id/relationships/vars/:var_id``

        `Variable Sets Update Variable API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#update-a-variable-in-a-variable-set>`_

        `Update Variable In Variable Set Sample Payload \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#sample-payload-3>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}/relationships/vars/{var_id}"
        return self._update(url, payload)

    def delete_var_from_varset(self, varset_id, var_id):
        """
        ``DELETE varsets/:varset_id/relationships/vars/:var_id``

        `Variable Sets Delete Variable API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#delete-a-variable-in-a-variable-set>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}/relationships/vars/{var_id}"
        return self._destroy(url)

    def list_vars_in_varset(self, varset_id):
        """
        ``GET varsets/:varset_id/relationships/vars``

        `Variable Sets List In Variable Set API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#list-variables-in-a-variable-set>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}/relationships/vars"
        return self._list(url)

    # WORKSPACE RELATIONSHIPS
    def apply_varset_to_workspace(self, varset_id, payload):
        """
        ``POST varsets/:varset_id/relationships/workspaces``

        `Variable Sets Apply To Workspace API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#apply-variable-set-to-workspaces>`_

        `Apply Variable Set to Workspace Sample Payload \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#sample-payload-4>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}/relationships/workspaces"
        return self._post(url, data=payload)

    def remove_varset_from_workspace(self, varset_id, payload):
        """
        ``DELETE varsets/:varset_id/relationships/workspaces``

        `Variable Sets Remove From Workspace API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#remove-a-variable-set-from-workspaces>`_

        `Remove Variable Set From Workspace Sample Payload \
            <https://www.terraform.io/docs/cloud/api/variable-sets.html#sample-payload-5>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}/relationships/workspaces"
        self._destroy(url, data=payload)


    # PROJECT RELATIONSHIPS
    def apply_varset_to_project(self, varset_id, payload):
        """
        ``POST varsets/:varset_id/relationships/projects``

        `Variable Sets Apply To Project API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/variable-sets#apply-variable-set-to-projects>`_

        `Apply Variable Set to Project Sample Payload \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/variable-sets#sample-payload-6>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}/relationships/projects"
        return self._post(url, data=payload)

    def remove_varset_from_project(self, varset_id, payload):
        """
        ``DELETE varsets/:varset_id/relationships/projects``

        `Variable Sets Remove From Project API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/variable-sets#remove-a-variable-set-from-projects>`_

        `Remove Variable Set From Project Sample Payload \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/variable-sets#sample-payload-7>`_
        """
        url = f"{self._endpoint_base_url}/{varset_id}/relationships/projects"
        self._destroy(url, data=payload)



"""
Module for testing the Terraform Cloud API Endpoint: Workspace Resources.
"""

from .base import TestTFCBaseTestCase


class TestTFCWorkspaceResources(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Workspace Resources.
    """

    _unittest_name = "ws-rsrc"
    _endpoint_being_tested = "workspace_resources"

    def setUp(self):
        self._ws = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())
        self._ws_id = self._ws["data"]["id"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)

    def test_workspace_resources(self):
        """
        Test the Workspace Resources API endpoints.
        """

        listed_resources = self._api.workspace_resources.list(self._ws_id)["data"]
        self.assertEqual(len(listed_resources), 0)



"""
Module for Terraform Cloud API Endpoint: Workspace Resources.
"""

from .endpoint import TFCEndpoint

class TFCWorkspaceResources(TFCEndpoint):
    """
    `Workspace Resources API Docs \
        <https://www.terraform.io/cloud-docs/api-docs/workspace-resources>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/workspaces"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def list(self, workspace_id):
        """
        ``GET /workspaces/:workspace_id/resources``

        `Workspace Resources List API Doc Reference \
            <https://www.terraform.io/cloud-docs/api-docs/workspace-resources#list-workspace-resources>`_

        `Query Parameter(s) Details \
            <https://www.terraform.io/cloud-docs/api-docs/workspace-resources#query-parameters>`__
        """
        url = f"{self._endpoint_base_url}/{workspace_id}/resources/"
        return self._list(url)



"""
Module for testing the Terraform Cloud API Endpoint: Project Team Access.
"""

from .base import TestTFCBaseTestCase


class TestTFCProjectTeamAccess(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Project Team Access.
    """

    _unittest_name = "team-acc"
    _endpoint_being_tested = "team_access"

    def setUp(self):
        # Create a test team
        self._team = self._api.teams.create(
            self._get_team_create_payload())["data"]
        self._team_id = self._team["id"]

        self._project = self._api.projects.create(
            self._get_project_create_payload())["data"]
        self._project_id = self._project["id"]

        # Invite a test user to this org, will be removed after
        invite = self._api.org_memberships.invite(self._get_org_membership_invite_payload())
        self._org_membership_id = invite["data"]["id"]

        # Create a test workspace
        workspace = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())["data"]
        self._ws_id = workspace["id"]
        self._ws_name = workspace["attributes"]["name"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_name=self._ws_name)
        self._api.teams.destroy(self._team_id)
        self._api.projects.destroy(self._project_id)
        self._api.org_memberships.remove(self._org_membership_id)

    def test_project_team_access(self):
        """
        Test the Team Access API endpoints.
        """
        listed_proj_filters = [
            {
                "keys": ["project", "id"],
                "value": self._project_id
            }
        ]
        listed_proj_access = self._api.project_team_access.list(filters=listed_proj_filters)
        self.assertEqual(len(listed_proj_access["data"]), 0)

        # Create new project team access, confirm it has been created
        proj_team_access_payload = {
            "data": {
                "attributes": {
                    "access": "read"
                },
                "relationships": {
                    "project": {
                        "data": {
                            "type": "projects",
                            "id": self._project_id
                        }
                    },
                    "team": {
                        "data": {
                            "type": "teams",
                            "id": self._team_id
                        }
                    }
                },
                "type": "team-projects"
            }
        }
        proj_team_access = self._api.project_team_access.add_project_team_access(
            proj_team_access_payload)
        proj_team_access_id = proj_team_access["data"]["id"]

        # Show the newly created project team access, confirm the ID matches to the created one
        shown_proj_team_access = self._api.project_team_access.show(proj_team_access_id)
        self.assertEqual(shown_proj_team_access["data"]["id"], proj_team_access_id)

        # Update team access
        update_proj_team_access_payload = {
            "data": {
                "id": proj_team_access_id,
                "attributes": {
                    "access": "admin"
                }
            }
        }
        updated_proj_team_access = \
            self._api.project_team_access.update(proj_team_access_id, update_proj_team_access_payload)["data"]
        self.assertEqual(updated_proj_team_access["attributes"]["access"], "admin")

        # Remove the team access, confirm it's gone
        self._api.project_team_access.remove_project_team_access(proj_team_access_id)
        listed_proj_access = self._api.project_team_access.list(filters=listed_proj_filters)
        self.assertEqual(len(listed_proj_access["data"]), 0)




"""
Module for Terraform Cloud API Endpoint: Projects.
"""

from .endpoint import TFCEndpoint

class TFCProjects(TFCEndpoint):
    """
    `Projects API Docs \
        <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/projects>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._projects_api_v2_base_url = f"{self._api_v2_base_url}/projects"
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/organizations/{org_name}/projects"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, payload):
        """
        ``POST /organizations/:organization_name/projects``

        `Projects Create API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/projects#create-a-project>`_

        `Create Sample Payload \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/projects#sample-payload>`_
        """
        return self._create(self._org_api_v2_base_url, payload)

    def destroy(self, project_id):
        """
        ``DELETE /projects/:project_id``

        `Projects Destroy API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/projects#delete-a-project>`_
        """
        url = f"{self._projects_api_v2_base_url}/{project_id}"
        return self._destroy(url)

    def list(self, page=None, page_size=None, filters=None, query=None):
        """
        ``GET /organizations/:organization_name/projects``

        `Projects List API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/projects#list-projects>`_

        `Query Parameter(s) Details \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/projects#query-parameters>`__
        """
        return self._list(\
            self._org_api_v2_base_url, page=page, page_size=page_size, filters=filters, query=query)

    def list_all(self, filters=None, query=None):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every run for a workspace.

        Returns an object with two arrays of objects.
        """
        return self._list_all(self._org_api_v2_base_url, filters=filters, query=query)

    def show(self, project_id):
        """
        ``GET /projects/:project_id``

        `Projects Show API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/projects#show-project>`_
        """
        url = f"{self._projects_api_v2_base_url}/{project_id}"
        return self._show(url)

    def update(self, project_id, payload):
        """
        ``PATCH /projects/:project_id``

        `Projects Update API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/projects#update-a-project>`_

        `Update Sample Payload \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/projects#sample-payload-1>`_
        """
        url = f"{self._projects_api_v2_base_url}/{project_id}"
        return self._update(url, payload)



"""
Module for Terraform Cloud API Endpoint: Comments.
"""

from .endpoint import TFCEndpoint

class TFCComments(TFCEndpoint):
    """
    `Comments API Docs \
        <https://www.terraform.io/cloud-docs/api-docs/comments>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/comments"
        self._runs_base_url = f"{self._api_v2_base_url}/runs"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def list_for_run(self, run_id):
        """
        ``GET /runs/:id/comments``

        `Comments List for Run API Doc Reference \
            <https://www.terraform.io/cloud-docs/api-docs/comments#list-comments-for-a-run>`_
        """
        url = f"{self._runs_base_url}/{run_id}/comments"
        return self._list(url)

    def show(self, comment_id):
        """
        ``GET /comments/:id``

        `Comments Show API Doc Reference \
            <https://www.terraform.io/cloud-docs/api-docs/comments#show-a-comment>`_
        """
        url = f"{self._endpoint_base_url}/{comment_id}"
        return self._show(url)

    def create_for_run(self, run_id, payload):
        """
        ``POST /runs/:id/comments``

        `Comments Create for Run API Doc Reference \
            <https://www.terraform.io/cloud-docs/api-docs/comments#create-comment>`_

        `Create Sample Payload \
            <https://www.terraform.io/cloud-docs/api-docs/comments#sample-payload>`_
        """
        url = f"{self._runs_base_url}/{run_id}/comments"
        return self._create(url, payload=payload)



"""
Module for testing the Terraform Cloud API Endpoint: GitHub Apps.
"""

from .base import TestTFCBaseTestCase


class TestTFCGitHubApps(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: GitHub Apps.
    """

    _unittest_name = "gitap"
    _endpoint_being_tested = "github_apps"

    def test_github_apps(self):
        """
        Test the GitHub Apps API endpoints.
        """
        listed_gh_apps = self._api.github_apps.list()["data"]
        gh_app_id = listed_gh_apps[0]["id"]
        self.assertTrue(len(listed_gh_apps) > 0)

        shown_gh_app = self._api.github_apps.show(gh_app_id)["data"]
        self.assertEqual(gh_app_id, shown_gh_app["id"])



"""
Module for testing the Terraform Cloud API Endpoint: Variable Sets.
"""

from .base import TestTFCBaseTestCase


class TestTFCVarSets(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Variable Sets.
    """

    _unittest_name = "vset"
    _endpoint_being_tested = "var_sets"

    def setUp(self):
        self._ws = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())
        self._ws_id = self._ws["data"]["id"]
        self._ws_name = self._ws["data"]["attributes"]["name"]
        self._variable_test_key = "terrasnek_unittest_key"
        self._variable_test_value = "terrasnek_unittest_value"

        self._project = self._api.projects.create(self._get_project_create_payload())
        self._project_id = self._project["data"]["id"]

    def tearDown(self):
        self._api.workspaces.destroy(
            workspace_name=self._ws["data"]["attributes"]["name"])

    def test_var_sets(self):
        """
        Test the Variable Sets API endpoints.
        """

        # List all the varsets in the org, confirm there are none
        listed_org_var_sets = self._api.var_sets.list_for_org()["data"]
        self.assertEqual(len(listed_org_var_sets), 0)

        all_listed_org_var_sets = self._api.var_sets.list_all_for_org()["data"]
        self.assertEqual(len(all_listed_org_var_sets), 0)

        # Setup a create payload that has no workspaces or variables attached.
        var_set_name = "terrasnek-unittest"
        create_payload = {
            "data": {
                "type": "varsets",
                "attributes": {
                    "name": var_set_name,
                    "description": "",
                    "is-global": False
                },
                "relationships": {
                    "workspaces": {
                        "data": [
                        ]
                    },
                    "vars": {
                        "data": [
                        ]
                    }
                }
            }
        }
        created_var_set = self._api.var_sets.create(create_payload)["data"]
        created_var_set_id = created_var_set["id"]
        self.assertEqual(var_set_name, created_var_set["attributes"]["name"])

        # Update the variable set and compare the names
        new_var_set_name = "terrasnek-unittest-updated"
        update_payload = {
            "data": {
                "type": "varsets",
                "attributes": {
                    "name": new_var_set_name,
                    "description": "",
                    "is-global": False
                },
                "relationships": {
                    "vars": {
                        "data": [
                        ]
                    }
                }
            }
        }
        updated_var_set = self._api.var_sets.update(created_var_set_id, update_payload)["data"]
        self.assertEqual(new_var_set_name, updated_var_set["attributes"]["name"])

        # Show the variable set and compare the names
        shown_var_set = self._api.var_sets.show(created_var_set_id)["data"]
        self.assertEqual(new_var_set_name, shown_var_set["attributes"]["name"])

        # Add a variable to the created variable set, confirm it's been added
        add_var_to_varset_payload = {
            "data": {
                "type": "vars",
                "attributes": {
                    "key": self._variable_test_key,
                    "value": self._variable_test_value,
                    "description": "foo",
                    "sensitive": False,
                    "category": "terraform",
                    "hcl": False
                }
            }
        }
        added_var = self._api.var_sets.add_var_to_varset(created_var_set_id, add_var_to_varset_payload)["data"]
        added_var_id = added_var["id"]
        self.assertEqual(added_var["attributes"]["key"], self._variable_test_key)

        # List the variables in the variable set, confirm the one that was just added is present
        listed_vars = self._api.var_sets.list_vars_in_varset(created_var_set_id)["data"]
        self.assertEqual(listed_vars[0]["attributes"]["key"], self._variable_test_key)

        # Update the value of the variable that was just added, confirm that the update took place
        updated_value = "foo"
        update_var_in_varset_payload = {
            "data": {
                "type": "vars",
                "attributes": {
                    "key": self._variable_test_key,
                    "value": updated_value,
                    "description": "new cheeeese",
                    "sensitive": False,
                    "category": "terraform",
                    "hcl": False
                }
            }
        }
        updated_var = self._api.var_sets.update_var_in_varset(\
            created_var_set_id, added_var_id, update_var_in_varset_payload)["data"]
        self.assertEqual(updated_var["attributes"]["value"], updated_value)

        # Apply the variable set to the workspace
        apply_varset_ws_payload = {
            "data": [
                {
                    "type": "workspaces",
                    "id": self._ws_id
                }
            ]
        }
        # There is no response when we apply a varset to a workspace
        self._api.var_sets.apply_varset_to_workspace(created_var_set_id, apply_varset_ws_payload)

        # Confirm the applied variable set is present on the workspace
        listed_workspace_varsets = self._api.var_sets.list_for_workspace(self._ws_id)["data"]
        self.assertEqual(listed_workspace_varsets[0]["id"], created_var_set_id)

        all_listed_workspace_varsets = self._api.var_sets.list_all_for_workspace(self._ws_id)["data"]
        self.assertEqual(all_listed_workspace_varsets[0]["id"], created_var_set_id)

        # Remove the variable set from the workspace
        remove_varset_payload = {
            "data": [
                {
                    "type": "workspaces",
                    "id": self._ws_id
                }
            ]
        }
        # There is no response when we remove a varset from a workspace
        self._api.var_sets.remove_varset_from_workspace(created_var_set_id, remove_varset_payload)

        # Confirm that there are no longer any varsets attached to the workspace
        listed_workspace_varsets = self._api.var_sets.list_for_workspace(self._ws_id)["data"]
        self.assertEqual(len(listed_workspace_varsets), 0)

        # Delete the variable added from the variable set and confirm it's gone
        self._api.var_sets.delete_var_from_varset(created_var_set_id, added_var_id)

        # Apply the variable set to a project
        varset_proj_payload = {
            "data": [
                {
                    "type": "projects",
                    "id": self._project_id
                }
            ]
        }
        self._api.var_sets.apply_varset_to_project(created_var_set_id, varset_proj_payload)
        # Confirm the variable set is attached to the project
        shown_var_set = self._api.var_sets.show(created_var_set_id)["data"]
        self.assertEqual(len(shown_var_set["relationships"]["projects"]["data"]), 1)
        # Remove the variable set from the project
        self._api.var_sets.remove_varset_from_project(created_var_set_id, varset_proj_payload)
        # Confirm the variable set is no longer attached to the project
        shown_var_set = self._api.var_sets.show(created_var_set_id)["data"]
        self.assertEqual(len(shown_var_set["relationships"]["projects"]["data"]), 0)

        # Destroy the variable set that was created earlier
        # There is no response when we destroy a variable set
        self._api.var_sets.destroy(created_var_set_id)

        # Confirm we no longer have any varsets
        listed_var_sets = self._api.var_sets.list_for_org()["data"]
        self.assertEqual(len(listed_var_sets), 0)



"""
Module for testing the Terraform Cloud API Endpoint: No Code Provisioning.
"""

from .base import TestTFCBaseTestCase
from ._constants import TFE_MODULE_PROVIDER_TYPE


class TestTFCNoCodeProvisioning(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: No Code Provisioning.
    """

    _unittest_name = "no-cde"
    _endpoint_being_tested = "no_code_provisioning"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        # Store the OAuth client ID to remove it at the end.
        oauth_client = self._api.oauth_clients.create(
            self._get_oauth_client_create_payload())["data"]
        self._oauth_client_id = oauth_client["id"]
        self._oauth_token_id = \
            oauth_client["relationships"]["oauth-tokens"]["data"][0]["id"]

    def tearDown(self):
        # Delete all the modules before deleting the OAuth client, otherwise
        # you might not be able to delete it after the oauth client is purged.
        self._purge_module_registry()
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_no_code_provisioning(self):
        """
        Test the No Code Provisioning API endpoints.
        """

        # Publish a module from the VCS provider, using the OAuth token generated
        # in the class setup. Assert that we got the expected response back.
        publish_payload = {
            "data": {
                "attributes": {
                    "vcs-repo": {
                        "identifier": "dahlke/terraform-tfe-terrasnek-unittest",
                        "oauth-token-id": self._oauth_token_id,
                        "display_identifier": "dahlke/terraform-tfe-terrasnek-unittest"
                    }
                },
                "type": "registry-modules"
            }
        }
        published_mod = \
            self._api.registry_modules.publish_from_vcs(publish_payload)["data"]
        published_mod_name = published_mod["attributes"]["name"]

        # Wait for the VCS and TFE to sync up and have "setup_complete" status on the module.
        self._setup_published_module_version_timeout(
            published_mod_name, TFE_MODULE_PROVIDER_TYPE)

        shown_reg_mod = \
            self._api.registry_modules.show(
                published_mod_name, TFE_MODULE_PROVIDER_TYPE)["data"]
        shown_reg_mod_id = shown_reg_mod["id"]
        latest_published_version = shown_reg_mod["attributes"]["version-statuses"][0]["version"]

        enable_payload = {
            "data": {
                "type": "no-code-modules",
                "attributes": {
                    "version-pin":  latest_published_version,
                    "enabled": True
                },
                "relationships": {
                    "registry-module": {
                        "data": {
                            "id": shown_reg_mod_id,
                            "type": "registry-module"
                        }
                    },
                    "variable-options": {
                        "data": []
                    }
                }
            }
        }

        # Enable no code provisioning for the module and check it is marked as enabled
        enabled_mod = self._api.no_code_provisioning.enable(enable_payload)["data"]
        enabled_mod_id = enabled_mod["id"]
        self.assertTrue(enabled_mod["attributes"]["enabled"])

        # Update the no code module, and confirm the updates took, reuse the enable_payload
        updated_mod = self._api.no_code_provisioning.update(enabled_mod_id, enable_payload)["data"]
        updated_mod_id = updated_mod["id"]
        self.assertEqual(updated_mod_id, enabled_mod_id)

        # Show the no code module, and confirm it matches the module we updated
        shown_mod = self._api.no_code_provisioning.show(enabled_mod_id)["data"]
        shown_mod_id = shown_mod["id"]
        self.assertEqual(shown_mod_id, updated_mod_id)



"""
Module for Terraform Cloud API Endpoint: Run Tasks Integration.
"""

from .endpoint import TFCEndpoint

class TFCRunTasksIntegration(TFCEndpoint):
    """
    `Run Tasks Integration API Docs \
        <https://www.terraform.io/cloud-docs/api-docs/run-tasks-integration>`_
    """

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def request(self, url):
        """
        ``POST :url``

        `Run Task Integration Request API Doc Reference \
            <https://www.terraform.io/cloud-docs/api-docs/run-tasks-integration#run-task-request>`_

        `Request Sample Payload \
            <https://www.terraform.io/cloud-docs/api-docs/run-tasks-integration#sample-payload>`_
        """
        return self._post(url)

    def callback(self, callback_url):
        """
        ``PATCH :callback_url``

        `Run Task Integration Callback API Doc Reference \
            <https://www.terraform.io/cloud-docs/api-docs/run-tasks-integration#run-task-callback>`_

        `Callback Sample Payload \
            <https://www.terraform.io/cloud-docs/api-docs/run-tasks-integration#request-body-1>`_
        """
        return self._patch(callback_url)



"""
Module for testing the Terraform Cloud API Endpoint: Run Tasks Integration.
"""

from .base import TestTFCBaseTestCase


class TestTFCRunTaskIntegration(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Run Tasks Integration.
    """

    _unittest_name = "run-tint"
    _endpoint_being_tested = "run_tasks_integration"

    def setUp(self):
        self._ws = self._api.workspaces.create(self._get_ws_no_vcs_create_payload())
        self._ws_id = self._ws["data"]["id"]
        self._ws_name = self._ws["data"]["attributes"]["name"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_name=self._ws_name)

    def test_run_tasks(self):
        """
        Test the Run Tasks Integration API endpoints.
        """
        # TODO: These will be fairly difficult to create valuable tests for.



"""
Module for Terraform Cloud API Endpoint: No Code Provisioning.
"""

from .endpoint import TFCEndpoint


class TFCNoCodeProvisioning(TFCEndpoint):
    """
    `No Code Provisioning API Docs \
        <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/no-code-provisioning>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers,
                         well_known_paths, verify, log_level)
        self._org_base_url = \
            f"{self._api_v2_base_url}/organizations/{org_name}/no-code-modules"
        self._no_code_base_url = \
            f"{self._api_v2_base_url}/no-code-modules"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def enable(self, payload):
        """
        ``POST /organizations/:organization_name/no-code-modules``

        `No Code Provisioning Enable API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/no-code-provisioning#allow-no-code-provisioning-of-a-module-within-an-organization>`_

        `Enable Sample Payload \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/no-code-provisioning#sample-payload>`_
        """
        return self._post(self._org_base_url, data=payload)

    def update(self, module_id, payload):
        """
        ``PATCH /no-code-modules/:id``

        `No Code Modules Update API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/no-code-provisioning#update-no-code-provisioning-settings-for-a-module>`_

        `Update Sample Payload \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/no-code-provisioning#sample-payload-1>`_
        """
        url = f"{self._no_code_base_url}/{module_id}"
        return self._patch(url, payload)

    def show(self, module_id):
        """
        ``GET /no-code-modules/:id``

        `No Code Modules Read Properties API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/no-code-provisioning#read-a-no-code-module-s-properties>`_
        """
        url = f"{self._no_code_base_url}/{module_id}"
        return self._show(url)



"""
Module for Terraform Cloud API Endpoint: Assessment Results.
"""

from .endpoint import TFCEndpoint

class TFCAssessmentResults(TFCEndpoint):
    """
    `Assessment Results API Docs \
        <https://www.terraform.io/cloud-docs/api-docs/assessment-results>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/assessment-results"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def show(self, assessment_result_id):
        """
        ``GET api/v2/assessment-results/:assessment_result_id``

        `Show Assessment Results API Doc Reference \
            <https://www.terraform.io/cloud-docs/api-docs/assessment-results#show-assessment-result>`_
        """
        url = f"{self._endpoint_base_url}/{assessment_result_id}"
        return self._show(url)

    def get_json_plan(self, assessment_result_id):
        """
        ``GET api/v2/assessment-results/:assessment_result_id/json-output``

        `Show Assessment Results JSON Plan API Doc Reference \
            <https://www.terraform.io/cloud-docs/api-docs/assessment-results#json-plan>`_
        """
        url = f"{self._endpoint_base_url}/{assessment_result_id}/json-output"
        return self._get(url)

    def get_json_schema(self, assessment_result_id):
        """
        ``GET api/v2/assessment-results/:assessment_result_id/json-schema``

        `Show Assessment Results JSON Schema API Doc Reference \
            <https://www.terraform.io/cloud-docs/api-docs/assessment-results#json-schema-file>`_
        """
        url = f"{self._endpoint_base_url}/{assessment_result_id}/json-schema"
        return self._get(url)

    def get_json_log_output(self, assessment_result_id):
        """
        ``GET api/v2/assessment-results/assessment_result_id/log-output``

        `Show Assessment Results JSON Log Output API Doc Reference \
            <https://www.terraform.io/cloud-docs/api-docs/assessment-results#json-log-output>`_
        """
        url = f"{self._endpoint_base_url}/{assessment_result_id}/log-output"
        return self._get(url)



"""
Module for testing the Terraform Cloud API Endpoint: Projects.
"""

from .base import TestTFCBaseTestCase


class TestTFCProjects(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Projects.
    """

    _unittest_name = "projs"
    _endpoint_being_tested = "projects"

    def test_projects(self):
        """
        Test the Projects API endpoints.
        """

        # List all the projects, confirm that the response type
        projects = self._api.projects.list()["data"]
        self.assertEqual("projects", projects[0]["type"])

        # Create a new project, confirm that it has been created
        new_project = self._api.projects.create(
            self._get_project_create_payload())["data"]
        new_project_id = new_project["id"]
        new_project_name = new_project["attributes"]["name"]

        # Confirm we have the default project and the newly created one
        some_projects_raw = self._api.projects.list()["data"]
        self.assertEqual("Default Project", some_projects_raw[0]["attributes"]["name"])
        self.assertEqual(new_project_id, some_projects_raw[-1]["id"])

        # Show the newly created project
        shown_project = self._api.projects.show(new_project_id)["data"]
        self.assertEqual(new_project_id, shown_project["id"])

        # TODO: test the filters param once permissions API is in place for terrasnek
        all_projects_raw = self._api.projects.list()["data"]
        self.assertEqual("Default Project", all_projects_raw[0]["attributes"]["name"])
        self.assertEqual(new_project_id, all_projects_raw[-1]["id"])

        # Update the project to have VCS management access, confirm the changes took effect.
        new_project_name = self._random_name(ran_str_len=4)
        update_payload = {
            "data": {
                "type": "projects",
                "attributes": {
                    "name": new_project_name
                }
            }
        }
        updated_project = self._api.projects.update(new_project_id, update_payload)["data"]
        self.assertEqual(new_project_name, updated_project["attributes"]["name"])

        # Destroy the project, confirm it's gone
        self._api.projects.destroy(new_project_id)
        some_projects = self._api.projects.list()["data"]
        found_project = False
        for project in some_projects:
            if project["id"] == new_project_id:
                found_project = True
                break
        self.assertFalse(found_project)



"""
Module for Terraform Cloud API Endpoint: Registry Modules.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements


class TFCRegistryModules(TFCEndpoint):
    """
    `Registry Modules API Docs (Private Registry) \
        <https://www.terraform.io/docs/cloud/api/modules.html>`_

    `Registry Modules API Docs (Public Registry) \
        <https://www.terraform.io/docs/registry/api.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        # Helper vars to help clarify how the URLs translate to docs below
        self._namespace = self._org_name
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/organizations"
        self._mods_v2_base_url = f"{self._org_api_v2_base_url}/{self._org_name}/registry-modules"
        """
        https://app.terraform.io/api/registry/v1/modules/terrasnek-unittest/terrasnek-unittest/tfe/versions
        True
        /api/v2/organizations/terrasnek-unittest/registry-modules/private/terrasnek-unittest/terrasnek-unittest/tfe
        {self._org_api_v2_base_url}/{reg_name}/{self._org_name}/{MOD NAME}/{PROVIDER}
        """

    def required_entitlements(self):
        return [Entitlements.PRIVATE_MODULE_REGISTRY]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    # Public Registry API Endpoints
    def list(self, offset=None, page=None, page_size=None, limit=None, \
        provider=None, verified=None, filters=None):
        """
        ``GET /organizations/:organization_name/registry-modules``

        `Registry Modules List API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#list-modules>`_

        `Query Parameter(s) Details \
            <https://www.terraform.io/docs/cloud/api/modules.html#query-parameters>`__
        """
        return self._list(\
            self._mods_v2_base_url, page=page, page_size=page_size, offset=offset, \
                limit=limit, provider=provider, verified=verified, filters=filters)

    def list_all(self):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every audit trail in an organization.

        Returns an object with two arrays of objects.
        """
        return self._list_all(self._mods_v2_base_url)


    def search(self, query, offset=None, limit=None, provider=None, verified=None):
        """
        ``GET <base_url>/search``

        `Registry Modules Search API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#search-modules>`_
        """
        url = f"{self._mods_v1_base_url}/search"
        return self._list(url, \
            query=query, offset=offset, limit=limit, provider=provider,\
            verified=verified)

    def show(self, module_name, provider, reg_name="private"):
        """
        ``GET /organizations/:organization_name/registry-modules/:registry_name/:namespace/:name/:provider``

        **NOTE: Deprecated**
        ``GET /registry-modules/show/:organization_name/:name/:provider``

        `Registry Modules Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#show-a-module>`_
        """
        # NOTE: Future - this may be modded to support non-private registries.
        # NOTE: File a GH issue if you need non-private registry support.
        url = f"{self._mods_v2_base_url}/{reg_name}/{self._namespace}/{module_name}/{provider}"

        return self._show(url)

    def list_versions(self, name, provider): # reg_name="private"
        """
        ``GET <base_url>/:namespace/:name/:provider/versions``

        `Registry Modules List Versions API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#list-available-versions-for-a-specific-module>`_
        """
        url = f"{self._mods_v1_base_url}/{self._org_name}/{name}/{provider}/versions"
        return self._get(url)

    def list_latest_version_all_providers(self, name, offset=None, limit=None):
        """
        ``GET <base_url>/:namespace/:name``

        `Registry Modules List Latest Version All Providers API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#list-latest-version-of-module-for-all-providers>`_
        """
        url = f"{self._mods_v1_base_url}/{self._org_name}/{name}"
        return self._list(url, offset=offset, limit=limit)

    def list_latest_version_specific_provider(self, name, provider):
        """
        ``GET <base_url>/:namespace/:name/:provider``

        `Registry Modules List Latest Version Specific Provider API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#latest-version-for-a-specific-module-provider>`_
        """
        url = f"{self._mods_v1_base_url}/{self._org_name}/{name}/{provider}"
        return self._get(url)

    def get(self, name, provider, version):
        """
        ``GET <base_url>/:namespace/:name/:provider/:version``

        `Registry Modules Get API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#get-a-specific-module>`_
        """
        url = f"{self._mods_v1_base_url}/{self._org_name}/{name}/{provider}/{version}"
        return self._get(url)

    def download_version_source(self, name, provider, version, target_path):
        """
        ``GET <base_url>/:namespace/:name/:provider/:version/download``

        `Registry Modules Download Version Source API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#download-source-code-for-a-specific-module-version>`_
        """
        url = f"{self._mods_v1_base_url}/{self._org_name}/{name}/{provider}/{version}/download"
        return self._download(url, target_path, header_with_url="X-Terraform-Get")

    def download_latest_source(self, name, provider, target_path):
        """
        ``GET <base_url>/:namespace/:name/:provider/download``

        `Registry Modules Download Latest Source API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#download-the-latest-version-of-a-module>`_
        """
        url = f"{self._mods_v1_base_url}/{self._org_name}/{name}/{provider}/download"
        return self._download(\
            url, target_path, header_with_url="X-Terraform-Get", allow_redirects=True)

    # Private Registry API Endpoints
    def publish_from_vcs(self, payload):
        """
        ``POST /organizations/:organization_name/registry-modules/vcs``

        `Registry Modules Publish a Private Module From VCS API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#publish-a-private-module-from-a-vcs>`_

        `Publish From VCS Sample Payload \
            <https://www.terraform.io/docs/cloud/api/modules.html#sample-payload>`_
        """
        with_vcs_url = f"{self._mods_v2_base_url}/vcs"
        return self._post(with_vcs_url, data=payload)

    def destroy(self, module_name, reg_name="private", provider=None, version=None):
        """
        ``DELETE /organizations/:organization_name/registry-modules/:registry_name/:namespace/:name/:provider/:version``
        ``DELETE /organizations/:organization_name/registry-modules/:registry_name/:namespace/:name/:provider``
        ``DELETE /organizations/:organization_name/registry-modules/:registry_name/:namespace/:name``

        **NOTE: Deprecated**
        ``POST /registry-modules/actions/delete/:organization_name/:name/:provider/:version``
        ``POST /registry-modules/actions/delete/:organization_name/:name/:provider``
        ``POST /registry-modules/actions/delete/:organization_name/:name``

        `Registry Modules Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#delete-a-module>`_
        """
        # NOTE: Namespace param is the org name for a private registry, and we only support private registries (for now)
        url = f"{self._mods_v2_base_url}/{reg_name}/{self._namespace}/{module_name}/"

        if provider:
            url += f"/{provider}"
            if version:
                url += f"/{version}"

        return self._destroy(url)

    def create(self, payload):
        """
        ``POST /organizations/:organization_name/registry-modules``

        `Registry Modules Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#create-a-module>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/modules.html#request-body-1>`_
        """

        url = f"{self._org_api_v2_base_url}/{self._org_name}/registry-modules"
        return self._post(url, data=payload)


    def create_version(self, module_name, provider, payload, reg_name="private"):
        """
        ``POST /organizations/:organization_name/registry-modules/:registry_name/:namespace/:name/:provider/versions``

        `Registry Modules Create Version API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#create-a-module-version>`_

        `Create Version Sample Payload \
            <https://www.terraform.io/docs/cloud/api/modules.html#request-body-2>`_
        """
        # NOTE: In the future, this may be modified to support non-private registries. File an issue if you need this.
        url = f"{self._mods_v2_base_url}/{reg_name}/{self._namespace}/{module_name}/{provider}/versions"
        return self._post(url, data=payload)

    def upload_version(self, path_to_tarball, upload_url):
        """
        ``PUT https://archivist.terraform.io/v1/object/<UNIQUE OBJECT ID>``

        `Registry Modules Upload Version API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#upload-a-module-version>`_
        """
        data = None

        with open(path_to_tarball, 'rb') as data_bytes:
            data = data_bytes.read()

        return self._put(upload_url, data=data)



"""
Module for Terraform Cloud API Endpoint: GitHub Apps.
"""

from .endpoint import TFCEndpoint

class TFCGitHubApps(TFCEndpoint):
    """
    `GitHub Apps API Docs \
        <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/github-app-installations>`_
    """

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def list(self, filters=None):
        """
        ``GET /github-app/installations``

        `GitHub Apps List API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/github-app-installations#list-installations>`_
        """
        url = f"{self._api_v2_base_url}/github-app/installations"
        return self._list(url, filters=filters)

    def show(self, app_install_id):
        """
        ``GET /github-app/installation/:gh_app_installation_id``

        `GitHub Apps Show API Doc Reference \
            <https://developer.hashicorp.com/terraform/cloud-docs/api-docs/github-app-installations#show-installation>`_
        """
        url = f"{self._api_v2_base_url}/github-app/installation/{app_install_id}"
        return self._list(url)
